@@ -3,7 +3,6 @@ use {super::*, serde::Serialize};
 #[derive(Debug)]
 struct Invocation<'src: 'run, 'run> {
   arguments: Vec<&'run str>,
-  module: &'run Justfile<'src>,
   recipe: &'run Recipe<'src>,
 }
 
@@ -14,7 +13,7 @@ pub(crate) struct Justfile<'src> {
   #[serde(rename = "first", serialize_with = "keyed::serialize_option")]
   pub(crate) default: Option<Arc<Recipe<'src>>>,
   pub(crate) doc: Option<String>,
-  pub(crate) groups: Vec<String>,
+  pub(crate) groups: Vec<StringLiteral<'src>>,
   #[serde(skip)]
   pub(crate) loaded: Vec<PathBuf>,
   #[serde(skip)]
@@ -50,12 +49,22 @@ impl<'src> Justfile<'src> {
       input,
       self
         .recipes
-        .keys()
-        .map(|name| Suggestion { name, target: None })
-        .chain(self.aliases.iter().map(|(name, alias)| Suggestion {
-          name,
-          target: Some(alias.target.name.lexeme()),
-        })),
+        .values()
+        .filter(|recipe| recipe.is_public())
+        .map(|recipe| Suggestion {
+          name: recipe.name(),
+          target: None,
+        })
+        .chain(
+          self
+            .aliases
+            .values()
+            .filter(|alias| alias.is_public())
+            .map(|alias| Suggestion {
+              name: alias.name.lexeme(),
+              target: Some(alias.target.name.lexeme()),
+            }),
+        ),
     )
   }
 
@@ -69,24 +78,23 @@ impl<'src> Justfile<'src> {
     )
   }
 
-  fn evaluate_assignments<'run>(
+  fn evaluate_scopes<'run>(
     &'run self,
     arena: &'run Arena<Scope<'src, 'run>>,
     config: &'run Config,
     dotenv: &'run BTreeMap<String, String>,
     overrides: &BTreeMap<String, String>,
     root: &'run Scope<'src, 'run>,
-    scopes: &mut BTreeMap<String, &'run Scope<'src, 'run>>,
+    scopes: &mut BTreeMap<String, (&'run Justfile<'src>, &'run Scope<'src, 'run>)>,
     search: &'run Search,
   ) -> RunResult<'src> {
     let scope = Evaluator::evaluate_assignments(config, dotenv, self, overrides, root, search)?;
 
     let scope = arena.alloc(scope);
-    scopes.insert(self.module_path.clone(), scope);
-    scopes.get(&self.module_path).unwrap();
+    scopes.insert(self.module_path.clone(), (self, scope));
 
     for module in self.modules.values() {
-      module.evaluate_assignments(
+      module.evaluate_scopes(
         arena,
         config,
         dotenv,
@@ -128,7 +136,7 @@ impl<'src> Justfile<'src> {
     let root = Scope::root();
     let arena = Arena::new();
     let mut scopes = BTreeMap::new();
-    self.evaluate_assignments(
+    self.evaluate_scopes(
       &arena,
       config,
       &dotenv,
@@ -138,7 +146,7 @@ impl<'src> Justfile<'src> {
       search,
     )?;
 
-    let scope = scopes.get(&self.module_path).unwrap();
+    let scope = scopes.get(&self.module_path).unwrap().1;
 
     match &config.subcommand {
       Subcommand::Command {
@@ -230,13 +238,6 @@ impl<'src> Justfile<'src> {
 
     let ran = Ran::default();
     for invocation in invocations {
-      let context = ExecutionContext {
-        config,
-        dotenv: &dotenv,
-        module: invocation.module,
-        search,
-      };
-
       Self::run_recipe(
         &invocation
           .arguments
@@ -244,11 +245,13 @@ impl<'src> Justfile<'src> {
           .copied()
           .map(str::to_string)
           .collect::<Vec<String>>(),
-        &context,
+        config,
+        &dotenv,
         false,
         &ran,
         invocation.recipe,
         &scopes,
+        search,
       )?;
     }
 
@@ -289,7 +292,6 @@ impl<'src> Justfile<'src> {
       let recipe = self.get_recipe(&path[position]).unwrap();
       Ok(Invocation {
         arguments: arguments.into(),
-        module: self,
         recipe,
       })
     } else {
@@ -308,13 +310,15 @@ impl<'src> Justfile<'src> {
 
   fn run_recipe(
     arguments: &[String],
-    context: &ExecutionContext<'src, '_>,
+    config: &Config,
+    dotenv: &BTreeMap<String, String>,
     is_dependency: bool,
-    ran: &Ran<'src>,
+    ran: &Ran,
     recipe: &Recipe<'src>,
-    scopes: &BTreeMap<String, &Scope<'src, '_>>,
+    scopes: &BTreeMap<String, (&Justfile<'src>, &Scope<'src, '_>)>,
+    search: &Search,
   ) -> RunResult<'src> {
-    let mutex = ran.mutex(&recipe.namepath, arguments);
+    let mutex = ran.mutex(recipe, arguments);
 
     let mut guard = mutex.lock().unwrap();
 
@@ -322,41 +326,59 @@ impl<'src> Justfile<'src> {
       return Ok(());
     }
 
-    if !context.config.yes && !recipe.confirm()? {
+    if !config.yes && !recipe.confirm()? {
       return Err(Error::NotConfirmed {
         recipe: recipe.name(),
       });
     }
 
-    let scope = scopes
-      .get(&recipe.module_path())
+    let (module, scope) = scopes
+      .get(recipe.module_path())
       .expect("failed to retrieve scope for module");
 
-    let (outer, positional) =
-      Evaluator::evaluate_parameters(context, is_dependency, arguments, &recipe.parameters, scope)?;
+    let context = ExecutionContext {
+      config,
+      dotenv,
+      module,
+      search,
+    };
+
+    let (outer, positional) = Evaluator::evaluate_parameters(
+      &context,
+      is_dependency,
+      arguments,
+      &recipe.parameters,
+      scope,
+    )?;
 
     let scope = outer.child();
 
-    let mut evaluator = Evaluator::new(context, true, &scope);
+    let mut evaluator = Evaluator::new(&context, true, &scope);
 
     Self::run_dependencies(
-      context,
+      config,
+      &context,
       recipe.priors(),
+      dotenv,
       &mut evaluator,
       ran,
       recipe,
       scopes,
+      search,
     )?;
 
-    recipe.run(context, &scope, &positional, is_dependency)?;
+    recipe.run(&context, &scope, &positional, is_dependency)?;
 
     Self::run_dependencies(
-      context,
+      config,
+      &context,
       recipe.subsequents(),
+      dotenv,
       &mut evaluator,
       &Ran::default(),
       recipe,
       scopes,
+      search,
     )?;
 
     *guard = true;
@@ -365,12 +387,15 @@ impl<'src> Justfile<'src> {
   }
 
   fn run_dependencies<'run>(
+    config: &Config,
     context: &ExecutionContext<'src, 'run>,
     dependencies: &[Dependency<'src>],
+    dotenv: &BTreeMap<String, String>,
     evaluator: &mut Evaluator<'src, 'run>,
-    ran: &Ran<'src>,
+    ran: &Ran,
     recipe: &Recipe<'src>,
-    scopes: &BTreeMap<String, &Scope<'src, 'run>>,
+    scopes: &BTreeMap<String, (&Justfile<'src>, &Scope<'src, 'run>)>,
+    search: &Search,
   ) -> RunResult<'src> {
     if context.config.no_dependencies {
       return Ok(());
@@ -389,10 +414,11 @@ impl<'src> Justfile<'src> {
       thread::scope::<_, RunResult>(|thread_scope| {
         let mut handles = Vec::new();
         for (recipe, arguments) in evaluated {
-          handles.push(
-            thread_scope
-              .spawn(move || Self::run_recipe(&arguments, context, true, ran, recipe, scopes)),
-          );
+          handles.push(thread_scope.spawn(move || {
+            Self::run_recipe(
+              &arguments, config, dotenv, true, ran, recipe, scopes, search,
+            )
+          }));
         }
         for handle in handles {
           handle
@@ -403,7 +429,9 @@ impl<'src> Justfile<'src> {
       })?;
     } else {
       for (recipe, arguments) in evaluated {
-        Self::run_recipe(&arguments, context, true, ran, recipe, scopes)?;
+        Self::run_recipe(
+          &arguments, config, dotenv, true, ran, recipe, scopes, search,
+        )?;
       }
     }
 
@@ -440,8 +468,12 @@ impl<'src> Justfile<'src> {
     recipes
   }
 
-  pub(crate) fn groups(&self) -> &[String] {
-    &self.groups
+  pub(crate) fn groups(&self) -> Vec<&str> {
+    self
+      .groups
+      .iter()
+      .map(|group| group.cooked.as_str())
+      .collect()
   }
 
   pub(crate) fn public_groups(&self, config: &Config) -> Vec<String> {
